.. _distortion_selectors:

Reference Selection Keyword Attributes for DISTORTION
+++++++++++++++++++++++++++++++++++++++++++++++++++++
CRDS automatically selects the appropriate DISTORTION reference
files based on the following keyword attributes.

N.B.: DISTORTION is not applicable for instruments not in the table.

========== ===================================================
Instrument  Keyword Attributes
========== ===================================================
WFI	    instrument, detector, optical_element, date, time
========== ===================================================
