Reference Selection Keywords for LINEARITY
++++++++++++++++++++++++++++++++++++++++++
CRDS selects appropriate LINEARITY references based on the following keyword
attributes.
All keyword attributes used for file selection are *required*.

========== =====================================
Instrument  Keyword Attributes
========== =====================================
WFI         instrument, detector, date, time
========== =====================================
