Type Specific Keyword Attributes for DARK
+++++++++++++++++++++++++++++++++++++++++
In addition to the standard reference file keyword attributes listed above,
the following keyword attributes are *required* in DARK reference files,
because they are used as CRDS selectors
(see :ref:`dark_selectors`):

===============   ======================================  ==============
Attribute           Fully qualified path                   Instruments
===============   ======================================  ==============
detector           model.meta.instrument.detector          WFI
===============   ======================================  ==============
