Reference Selection Keywords for GAIN
+++++++++++++++++++++++++++++++++++++
CRDS selects appropriate GAIN references based on the following keywords.
GAIN is not applicable for instruments not in the table.
All keywords used for file selection are *required*.

========== ================================================
Instrument Keywords
========== ================================================
WFI        instrument, detector, date, time
========== ================================================
