"""Global configuration parameters"""

# Debug mode
#    In conjunction with logging, this mode changes the behavior
#    of functions. Common use case is where associations would be
#    removed, they are no longer removed. This assists in determining
#    why they were an issue.
DEBUG = False
